import warnings
from dataclasses import dataclass
from typing import Optional

import numpy as np
import pandas as pd
import pandas.api.types as pd_types
import plotly.express as px
import plotly.graph_objs as go
from scipy.stats import chisquare
from typing_extensions import Self

from ..statistical_testing import eval_ttest_significance, t_test
from .base import BaseEvaluator


@dataclass
class SRMCheckResult:
    variant: int
    sample_size: int
    sample_size_c: int
    chi_square: float
    p_value: float
    significant: bool
    segment: Optional[str] = None

    def __repr__(self) -> str:
        if self.segment:
            return f"variant: {self.variant}, segment: {self.segment} control:treatment = {self.sample_size_c}:{self.sample_size}, p_value = {self.p_value}"
        else:
            return f"variant: {self.variant} control:treatment = {self.sample_size_c}:{self.sample_size}, p_value = {self.p_value}"


class ABTestEvaluator(BaseEvaluator):
    def __init__(self) -> None:
        super().__init__()
        self.variant_col: str = ""
        self.segment_col: str = ""

    # ignore mypy error temporary, because the "Self" type support on mypy is ongoing. https://github.com/python/mypy/pull/11666
    def evaluate(
        self,
        data: pd.DataFrame,
        unit_col: str,
        metrics: list[str],
        variant_col: str = "variant",
        segment_col: Optional[str] = None,
    ) -> Self:  # type: ignore
        """calculate stats of A/B test and cache it into the class variable.
        At first, it only assumes metrics can handle by Welch's t-test.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe has randomization unit column, variant assignment column, and metrics columns.
            The data should have been aggregated by the randomization unit.
        unit_col : str
            A column name stores the randomization unit. something like user_id, session_id, ...
        metrics : list[str]
            Columns stores metrics you want to evaluate.
        variant_col : str
            A column name stores the variant assignment.
            The control variant should have value 1.
        segment_col : Optional[str]
            A column name stores 'segment' you want to break down in the analysis.
            e.g., light users, heavy users, new registers, ...
            When the specified column in dataframe stores numerical values, it automatically binning the values.

        Returns
        -------
        self : object
            Evaluator storing statistics calculated.
        """
        self._validate_passed_data(data, unit_col, metrics)

        # to avoid errors in later steps
        data[variant_col] = data[variant_col].astype(int)
        for metric_col in metrics:
            data[metric_col] = data[metric_col].astype(np.float64)

        if segment_col:
            segment = data[segment_col]
            if pd_types.is_numeric_dtype(segment) and not pd_types.is_bool_dtype(segment):
                segment = pd.qcut(x=segment, q=5, duplicates="drop")

            stats = pd.DataFrame()
            for s in segment.unique():
                partial_stats = t_test(data.loc[segment == s], unit_col, variant_col, metrics)
                partial_stats[segment_col] = s
                stats = pd.concat([stats, partial_stats])
            self.stats = stats
            self.segment_col = segment_col
        else:
            self.stats = t_test(data, unit_col, variant_col, metrics)
        self.variant_col = variant_col
        return self

    def summary_table(self, p_threshold: float = 0.05) -> pd.DataFrame:
        """return statistics summary.

        Parameters
        ----------
        p_threshold : float, optional
            significance level, by default 0.05

        Returns
        -------
        pd.DataFrame
            stats summary
        """
        self._validate_evaluate_executed()

        stats = self.stats.copy(deep=True)
        significance, abs_ci_width, rel_ci_width = eval_ttest_significance(self.stats, p_threshold)
        stats["significance"] = significance
        stats["abs_ci_width"] = abs_ci_width
        stats["rel_ci_width"] = rel_ci_width
        for diff_type in ["abs", "rel"]:
            stats[f"ci_{diff_type}_diff"] = stats.apply(
                lambda x: (
                    x[f"{diff_type}_diff_mean"] - x[f"{diff_type}_ci_width"],
                    x[f"{diff_type}_diff_mean"] + x[f"{diff_type}_ci_width"],
                ),
                axis=1,
            )
            del stats[f"{diff_type}_ci_width"]

        srm_check_results = self._diagnose_srm()
        for result in srm_check_results:
            if result.significant:
                print(f"[Warning] SRM was detected, pay attention to interpret the result. {result}")

        return_cols = [col for col in stats.columns if col[-2:] != "_c"]
        return stats.loc[:, return_cols]

    def summary_plot(self, p_threshold: float = 0.05, diff_type: str = "rel", display_ci: bool = True) -> go.Figure:
        """plot impact and confidence interval for each metric

        Parameters
        ----------
        p_threshold : float, optional
            significance level, by default 0.05
        diff_type : str, optional
            The type of difference you want to plot, by default 'rel'
        display_ci : bool, optional
            Whether to display confidence interval, by default True

        Returns
        -------
        go.Figure
        """
        self._validate_evaluate_executed()
        if diff_type not in ["rel", "abs"]:
            raise ValueError("Specified diff type is invalid.")

        stats = self.stats.copy(deep=True)
        significance, abs_ci_width, rel_ci_width = eval_ttest_significance(self.stats, p_threshold)
        stats["significant"] = significance
        stats["abs_ci_width"] = abs_ci_width
        stats["rel_ci_width"] = rel_ci_width

        viz_options = {
            "data_frame": stats.loc[stats[self.variant_col] > 1],  # not display control group
            "x": f"{diff_type}_diff_mean",
            "y": "metric",
            "facet_col": self.variant_col,
            "color": "significant",
            "color_discrete_map": {"up": "#54A24B", "down": "#E45756", "unclear": "silver"},
        }
        if display_ci:
            viz_options["error_x"] = f"{diff_type}_ci_width"
        if len(self.segment_col) > 0:
            viz_options["facet_row"] = self.segment_col
            viz_options["height"] = stats[self.segment_col].nunique() * 200

        srm_check_results = self._diagnose_srm()
        for result in srm_check_results:
            if result.significant:
                print(f"[Warning] SRM was detected, pay attention to interpret the result. {result}")

        g = px.bar(**viz_options)
        return g

    def summary_barplot(self, p_threshold: float = 0.05, diff_type: str = "rel", display_ci: bool = True) -> go.Figure:
        """plot impact and confidence interval for each metric

        Parameters
        ----------
        p_threshold : float, optional
            significance level, by default 0.05
        diff_type : str, optional
            The type of difference you want to plot, by default 'rel'
        display_ci : bool, optional
            Whether to display confidence interval, by default True

        Returns
        -------
        go.Figure
        """
        warnings.warn(
            "The summary_barplot() is deprecated and will be removed in the future release."
            "Please use summary_plot() instead.",
            FutureWarning,
        )
        return self.summary_plot(p_threshold=p_threshold, diff_type=diff_type, display_ci=display_ci)

    def _diagnose_srm(self) -> list[SRMCheckResult]:
        """Diagnosing Sample Ratio Mismatch by applying chi-square goodness of fit test.
        Assuming each variant has the same sample ratio.
        See this paper to understand what the SRM is: https://dl.acm.org/doi/10.1145/3292500.3330722

        Returns
        -------
        list[SRMCheckResult]
        """
        self._validate_evaluate_executed()
        if self.segment_col:
            stats_subset = self.stats[[self.variant_col, self.segment_col, "count", "count_c"]].drop_duplicates()
        else:
            stats_subset = self.stats[[self.variant_col, "count", "count_c"]].drop_duplicates()
        chi2q, p_value = chisquare(f_obs=stats_subset[["count", "count_c"]], axis=1)
        stats_subset["chi_square"] = chi2q
        stats_subset["p_value"] = p_value
        stats_subset["significant"] = p_value < 0.05
        stats_subset = stats_subset.loc[stats_subset[self.variant_col] > 1]

        results = []
        for _, row in stats_subset.iterrows():
            result = SRMCheckResult(
                variant=row[self.variant_col],
                sample_size=row["count"],
                sample_size_c=row["count_c"],
                chi_square=row["chi_square"],
                p_value=row["p_value"],
                significant=row["significant"],
            )
            if self.segment_col:
                result.segment = row[self.segment_col]
            results.append(result)
        return results



import numpy as np
import pandas as pd


def create_sample_biased(sample_size: int = 1000) -> pd.DataFrame:
    """Generate sample biased dataset for testing advanced causal inference approach. (e.g., linear regression, propensity score matching, ...)

    Parameters
    ----------
    sample_size : int, optional
        number of samples you want, by default 1000

    Returns
    -------
    pd.DataFrame
        Sample biased experiment result contains columns as follows
        - unit: something like user_id, session_id, ...
        - covar_numer1: dummy covariate variable (numerical)
        - covar_numer2: dummy covariate variable (numerical)
        - covar_cat: dummy covariate variable (categorical)
        - treatment: indicator variable that each unit received treatment or not
        - target: dummy target variable, interested in the treatment impact on
    """
    result = pd.DataFrame()
    result["unit"] = [i for i in range(sample_size)]
    result["covar_numer1"] = np.random.normal(size=sample_size)
    result["covar_numer2"] = np.random.exponential(size=sample_size)
    result["covar_cat"] = np.random.choice(a=[str(i + 1) for i in range(5)], size=sample_size)

    # generate probability of receiving treatment by subjective relationships
    relationship = (
        np.random.normal(loc=-1.0, scale=2.0, size=sample_size)
        + np.random.normal(loc=-2.0, scale=1.0, size=sample_size) * result["covar_numer1"]
        + np.random.normal(loc=5.0, scale=5.0, size=sample_size) * result["covar_numer2"]
    )
    treatment_prob = 1 / (1 + np.exp(-1.0 * relationship))
    result["treatment"] = (treatment_prob >= 0.5).astype("int")

    # generate target by subjective relationships
    result["target"] = (
        np.random.normal(loc=3.0, scale=5.0, size=sample_size)
        + np.random.normal(loc=1.0, scale=1.0, size=sample_size) * result["treatment"]
        + np.random.normal(loc=-0.5, scale=1.0, size=sample_size) * result["covar_numer1"]
        + np.random.normal(loc=2.0, scale=2.0, size=sample_size) * result["covar_numer2"]
    )

    return result



import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import statsmodels.formula.api as smf
from typing_extensions import Self

from .base import BaseEvaluator


class LinearRegressionEvaluator(BaseEvaluator):
    """Evaluate treatment impact by Linear Regression

    Attributes
    ----------
    models : dict[str, smf.ols]
        linear regression models built to evaluate the impact of each metric.
    """

    def __init__(self) -> None:
        super().__init__()
        self.models: dict[str, smf.ols] = dict()

    # ignore mypy error temporary, because the "Self" type support on mypy is ongoing. https://github.com/python/mypy/pull/11666
    def evaluate(
        self,
        data: pd.DataFrame,
        unit_col: str,
        metrics: list[str],
        treatment_col: str = "treatment",
        covariates: list[str] = [],
    ) -> Self:  # type: ignore
        """Evaluate impact by Linear Regression model

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe has randomization unit column, and metrics columns. The data should have been aggregated by the randomization unit.
        unit_col : str
            A column name stores the randomization unit. something like user_id, session_id, ...
        metrics : list[str]
            Columns stores metrics you want to evaluate.
        treatment_col : str
            A column name stores information if the unit received treatment or not.
            Stored value should be binary:
                - 0: did not receive treatment
                - 1: received treatment
        covariates : list[str], optional
            column names of covariate variables you want to use for the analysis, by default []

        Returns
        -------
        Self
            Evaluator object has statistics calculated
        """
        self._validate_passed_data(data, unit_col, metrics)
        if set(data[treatment_col].unique()) != {0, 1}:
            raise ValueError("The treatment value should be binary.")

        # to avoid errors in later steps
        data[treatment_col] = data[treatment_col].astype(int)
        for metric_col in metrics:
            data[metric_col] = data[metric_col].astype(np.float64)

        covariates_str = ""
        if len(covariates) > 0:
            covariates_str = "+ " + "+ ".join(covariates)

        stats = self.stats.copy(deep=True)
        for metric in metrics:
            model = smf.ols(formula=f"{metric} ~ {treatment_col} {covariates_str}", data=data).fit()
            self.models[metric] = model

            # convert statsmodels summary table to pandas dataframe
            stats_partial = pd.read_html(model.summary().tables[1].as_html(), header=0, index_col=0)[0].reset_index()
            stats_partial["metric"] = metric
            stats = pd.concat([stats, stats_partial])

        self.stats = stats.loc[stats["index"] == treatment_col].reset_index(drop=True).drop("index", axis=1)
        return self

    def summary_table(self) -> pd.DataFrame:
        self._validate_evaluate_executed()
        stats = self.stats.copy(deep=True)
        return stats

    def summary_plot(self, p_threshold: float = 0.05, display_ci: bool = True) -> go.Figure:
        """Plot evaluated impact and confidence interval for each metric.
        Currently it only supports the absolute difference visualization.

        Parameters
        ----------
        p_threshold : float, optional
            significance level, by default 0.05
        display_ci : bool, optional
            Whether to display confidence interval, by default True

        Returns
        -------
        go.Figure
        """
        stats = self.stats.copy(deep=True)
        self._validate_evaluate_executed()
        stats["abs_ci_width"] = stats["coef"] - stats["[0.025"]
        stats["significant"] = stats.apply(
            lambda x: (
                "up"
                if x["P>|t|"] <= p_threshold and x["t"] > 0
                else "down" if x["P>|t|"] <= p_threshold and x["t"] < 0 else "unclear"
            ),
            axis=1,
        )
        stats.rename(columns={"coef": "impact_abs"}, inplace=True)

        viz_options = {
            "data_frame": stats,
            "x": "impact_abs",
            "y": "metric",
            "color": "significant",
            "color_discrete_map": {"up": "#54A24B", "down": "#E45756", "unclear": "silver"},
        }
        if display_ci:
            viz_options["error_x"] = "abs_ci_width"

        g = px.bar(**viz_options)
        return g



from .aatest import AATestEvaluator
from .abtest import ABTestEvaluator
from .linear_regression import LinearRegressionEvaluator
from .samplesize import SampleSizeEvaluator

__all__ = ["ABTestEvaluator", "AATestEvaluator", "SampleSizeEvaluator", "LinearRegressionEvaluator"]



from typing import Optional

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from typing_extensions import Self

from .base import BaseEvaluator


class SampleSizeEvaluator(BaseEvaluator):
    def __init__(self) -> None:
        super().__init__()

    # ignore mypy error temporary, because the "Self" type support on mypy is ongoing. https://github.com/python/mypy/pull/11666
    def evaluate(self, data: pd.DataFrame, unit_col: str, metrics: list[str], n_variant: int = 2) -> Self:  # type: ignore
        """Calculate statistics of metrics and mde with simulating A/B test threshold.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe has randomization unit column, and metrics columns. The data should have been aggregated by the randomization unit.
        unit_col : str
            A column name stores the randomization unit. something like user_id, session_id, ...
        metrics : list[str]
            Columns stores metrics you want to evaluate.
        n_variant : int, optional
            The number of variant planned in the A/B test, by default 2

        Returns
        -------
        self : object
            Evaluator storing statistics calculated.
        """
        self._validate_passed_data(data, unit_col, metrics)

        # to avoid errors in later steps
        for metric_col in metrics:
            data[metric_col] = data[metric_col].astype(np.float64)

        stats = pd.DataFrame()
        for metric in metrics:
            stats_partial = pd.DataFrame()
            stats_partial["threshold"] = np.arange(0.01, 1.01, 0.01)
            stats_partial["metric"] = metric
            stats_partial["mean"] = data[metric].mean()
            stats_partial["var"] = data[metric].var()
            stats_partial["count"] = data[metric].count()
            stats = pd.concat([stats, stats_partial])

        stats["sample_size"] = stats["threshold"] * stats["count"] / n_variant
        stats["mde_abs"] = 4 * np.sqrt(stats["var"] / stats["sample_size"])
        stats["mde_rel"] = stats["mde_abs"] / stats["mean"]
        self.stats = stats
        return self

    def summary_table(self, target_mde: Optional[float] = None) -> pd.DataFrame:
        """Find threshold suffices the provided target MDE.

        Parameters
        ----------
        target_mde : Optional[float], optional
            The relative Minimum Detectable Effect you want to set in the A/B test.

        Returns
        -------
        pd.DataFrame
            stats summary
        """
        self._validate_evaluate_executed()
        _validate_target_mde(target_mde)
        stats = self.stats.copy(deep=True)
        if target_mde:
            return stats.loc[stats["mde_rel"] <= target_mde].groupby("metric").head(1)
        return stats

    def summary_plot(self, target_mde: Optional[float] = None) -> go.Figure:
        """Plot threshold vs MDE curve

        Parameters
        ----------
        target_mde : Optional[float], optional
            The relative Minimum Detectable Effect you want to set in the A/B test.

        Returns
        -------
        go.Figure
        """
        self._validate_evaluate_executed()
        _validate_target_mde(target_mde)

        g = px.line(data_frame=self.stats, x="threshold", y="mde_rel", color="metric", markers=True)
        if target_mde:
            g.add_hline(y=target_mde, line_dash="dot", line_width=2)
        return g


def _validate_target_mde(target_mde: Optional[float] = None) -> None:
    if target_mde is None or (0 < target_mde < 1):
        return
    raise ValueError("The target MDE should be (0, 1).")
